# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"get_aug_pipe": "30-swav.ipynb",
         "create_encoder": "30-swav.ipynb",
         "MLP": "30-swav.ipynb",
         "SimCLRModel": "10-simclr.ipynb",
         "create_simclr_model": "10-simclr.ipynb",
         "remove_diag": "10-simclr.ipynb",
         "SimCLRLoss": "10-simclr.ipynb",
         "SimCLR": "10-simclr.ipynb",
         "BYOLModel": "20-byol.ipynb",
         "create_byol_model": "20-byol.ipynb",
         "symmetric_mse_loss": "20-byol.ipynb",
         "byol_loss": "20-byol.ipynb",
         "BYOL": "20-byol.ipynb",
         "SwAVModel": "30-swav.ipynb",
         "create_swav_model": "30-swav.ipynb",
         "sinkhorn_knopp": "30-swav.ipynb",
         "SWAVLoss": "30-swav.ipynb",
         "SWAV": "30-swav.ipynb"}

modules = ["simclr.py",
           "byol.py",
           "swav.py"]

doc_url = "https://keremturgutlu.github.io/self_supervised/"

git_url = "https://github.com/keremturgutlu/self_supervised/tree/master/"

def custom_doc_links(name): return None
